/*

Casio Keyboard ROM Pack Emulator v1.5 - RO-000 "Test ROM"
  by Ra_226.  Jan 29, 2023

The ROM has a 4-bit bus and data is stored and addressed in 4-bit words.  There are 16-bits of address space, however only the top 12-bits are 
sent--the bottom 4 bits are implicitly zero, making data addressable only in 16-word blocks.  12 bits can address up to 4096 blocks.  The typical
6 kB (12k word) ROM has 768 addresses; the larger 24 kB (48k word) ROMs have 3072.

The ROM operates in one of two modes: ADDRESSING mode takes in an address and moves the pointer, while DATA OUT mode reads out to the bus.  For
an addressing operation, the CPU brings CE and OP Low, then sends 5 words for two clocks each: first the READ op code (0000), then three address
words (bits 4-7, bits 8-11, then bits 12-15), and finally the device code, which should be 0xC.  If the device code does not match, it's assumed
the CPU is communicating with another device and the ROM should not respond to any DATA OUT until a proper device code has been received via
addressing.  After addressing, CE and OP are brought high again and the ROM is given time to find the data (typically a dozen clocks or so).

For a DATA OUT operation, CE is brought low but OP remains high.  Assuming this follows a successful addressing op, the DATA lines switch from
high-impedance inputs to outputs and data is read out, one word every two clocks.  As long as CE is kept low, the ROM auto-increments the address
pointer, reading out data sequentially.  If CE is brought high, the ROM ceases outputting data, but the address pointer holds position.  If CE is
brought low again, it picks up where it left off unless a new addressing operation occurs.

You may notice the binary data from the ROM is stored "word-swapped."  That is, A5 00 00 D0 is actually output as 5A 00 00 0D, etc... It makes
the emulator slightly more complicated, but it means the code can stay in its more human-readable form, which I prefer (see my notes on the ROM
code analysis).  Most keyboards read one or more full bytes at a time, but the PT-50 (and possibly others) is an exception, reading only a single
word at a time.  The code must account for this behavior and pick up where it left off when these operations occur (see comments below).

A few more oddities of the Casio ROM:
- The power supply is negative, which means for power connections, GND is connected to 5V, and Vdd is 0V
- Since Vdd is considered HIGH, "negative logic" is in effect: 0V (really -5V) is HIGH and 5V (GND) is LOW (see comments below).
- The ROM uses a bi-phase clock where two clocks alternate back and forth, with a "dead" period between them when neither is active.
  - Due to negative logic, the clocks are considered active LOW (with a 25% duty cycle)
  - Signals from the keyboard (CE and OP changes, and data during ADDRESSING) occur on CK2
  - Signals from the ROM (data during DATA OUT) occur slightly after CK1, but always before the next CK2


Arduino Pin   Signal  ROM Pin     Notes
------------------------------------------------------------------------------------------------
PORTD
  Pin 2       CK1     Pin 5       Outputs (data during DATA OUT) appear on CK1
  Pin 3       CK2     Pin 6       Inputs (CE and OP, data during ADDRESSING) appear on CK2
  Pin 4       DATA1   Pin 8
  Pin 5       DATA2   Pin 9
  Pin 6       DATA3   Pin 10
  Pin 7       DATA4   Pin 11
PORTB
  Pin 10      OP      Pin 7       OP, CE High: ROM is idle
  Pin 11      CE      Pin 2       OP, CE Low: ADDRESSING mode.  Only CE Low: DATA OUT mode
  Pin 12      Select  --          Optional (possible future ROM select feature)
  Pin 13      Light   --          Optional (built-in LED, possible future ROM select indicator)
Power
  5V          GND     Pin 1       NC (would be 5V supply, but not enough power for Arduino)
  GND         VDD1    Pin 3       Connect to GND
  --          VDD2    Pin 4       NC (possibly used to put ROM in a power-saving state)
------------------------------------------------------------------------------------------------


How to Use:
- Use https://tomeko.net/online_tools/file_to_hex.php to convert a ROM .bin file to hex list
- Copy the results into the PROGMEM variable below
- Change dataROM[size] to 6144 or 24576 (or whatever size your ROM is, but these are typical)
- Upload to Emulator as you would any other Arduino code
- Code is intended for 5V, 16 MHz Arduino Pro Mini.  Not tested on 3.3V / 8MHz

*/

const PROGMEM byte dataROM[4096] = {
// 0     1     2     3     4     5     6     7     8     9     A     B     C     D     E     F    // RO-000 (4096 bytes)
0xA5, 0x00, 0x00, 0xD0, 0x1F, 0x00, 0x07, 0x00, 0x9C, 0x1F, 0x00, 0x00, 0x02, 0x00, 0x00, 0x06, 
0x00, 0x00, 0x0A, 0x00, 0x00, 0x0E, 0x00, 0x00, 0x12, 0x00, 0x00, 0x18, 0x00, 0x00, 0x1E, 0x00, 
0xA2, 0x1F, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x52, 0x4F, 0x2D, 0x30, 0x30, 0x30, 0x20, 0x00, 0x54, 0x65, 0x73, 0x74, 0x20, 0x52, 0x4F, 0x4D, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x62, 0x79, 0x20, 0x52, 0x61, 0x32, 0x32, 0x36, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x20, 0x02, 0x00, 0x10, 0x20, 0x04, 0x00, 0x02, 0x60, 0x04, 0x00, 0xFF, 0x88, 0x04, 0x00, 
0x10, 0x00, 0x00, 0x0E, 0x00, 0x00, 0x1E, 0x44, 0x00, 0x2E, 0x06, 0x00, 0x60, 0x00, 0x60, 0x0F, 
0x00, 0x00, 0x46, 0x18, 0x00, 0x48, 0x18, 0x00, 0x4A, 0x18, 0x00, 0x4B, 0x18, 0x00, 0x51, 0x18, 
0x00, 0x53, 0x18, 0x00, 0x55, 0x18, 0x00, 0x56, 0x18, 0x00, 0x58, 0x18, 0x00, 0x5A, 0x18, 0x00, 
0x5B, 0x18, 0x00, 0x61, 0x18, 0x00, 0x46, 0x18, 0x00, 0x45, 0x18, 0x00, 0x43, 0x18, 0x00, 0x41, 
0x18, 0x00, 0x3B, 0x18, 0x00, 0x3A, 0x18, 0x00, 0x38, 0x18, 0x00, 0x36, 0x60, 0x0C, 0x8F, 0x00, 
0x00, 0x60, 0x80, 0x00, 0x60, 0x10, 0x0C, 0x1F, 0x00, 0x00, 0x9F, 0x00, 0x00, 0x60, 0x90, 0x00, 
0x60, 0x20, 0x0C, 0x1F, 0x00, 0x00, 0xAF, 0x00, 0x00, 0x60, 0xA0, 0x00, 0x60, 0x30, 0x0C, 0x1F, 
0x00, 0x00, 0xBF, 0x00, 0x00, 0x60, 0xB0, 0x00, 0x60, 0x40, 0x0C, 0x0F, 0x00, 0x00, 0x46, 0x18, 
0x00, 0x48, 0x18, 0x00, 0x4A, 0x18, 0x00, 0x4B, 0x18, 0x00, 0x51, 0x18, 0x00, 0x53, 0x18, 0x00, 
0x55, 0x18, 0x00, 0x56, 0x18, 0x00, 0x58, 0x18, 0x00, 0x5A, 0x18, 0x00, 0x5B, 0x18, 0x00, 0x61, 
0x18, 0x00, 0x46, 0x18, 0x00, 0x45, 0x18, 0x00, 0x43, 0x18, 0x00, 0x41, 0x18, 0x00, 0x3B, 0x18, 
0x00, 0x3A, 0x18, 0x00, 0x38, 0x18, 0x00, 0x36, 0x60, 0x0C, 0x8F, 0x00, 0x00, 0x60, 0xC0, 0x00, 
0x60, 0x50, 0x0C, 0x1F, 0x00, 0x00, 0x9F, 0x00, 0x00, 0x60, 0xD0, 0x00, 0x60, 0x60, 0x0C, 0x1F, 
0x00, 0x00, 0xAF, 0x00, 0x00, 0x60, 0xE0, 0x00, 0x60, 0x70, 0x0C, 0x1F, 0x00, 0x00, 0xBF, 0x00, 
0x00, 0x60, 0xF0, 0x0C, 0xF0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x10, 0x00, 0x00, 0x60, 0x00, 0x00, 0x60, 0x80, 0x60, 0x10, 0x00, 0x12, 0xF0, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x10, 0x00, 0x90, 0x00, 0x80, 0x80, 0xC0, 0x2F, 0x10, 0x5D, 0x50, 0x80, 0xF1, 0x03, 0x20, 0x12, 
0xC0, 0xAF, 0xF0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x20, 0x06, 0x00, 0x10, 0x60, 0x06, 0x00, 0x02, 0x40, 0x08, 0x00, 0xFF, 0x68, 0x08, 0x00, 
0x10, 0x00, 0x00, 0x0E, 0x00, 0x00, 0x1E, 0x44, 0x00, 0x2E, 0x06, 0x00, 0x60, 0x00, 0x00, 0x60, 
0x80, 0x60, 0x10, 0x00, 0x12, 0xF0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x10, 0x00, 0x00, 0x60, 0x00, 0x60, 0x0F, 0x00, 0x00, 0x46, 0x18, 0x00, 0x48, 0x18, 0x00, 0x4A, 
0x18, 0x00, 0x4B, 0x18, 0x00, 0x51, 0x18, 0x00, 0x53, 0x18, 0x00, 0x55, 0x18, 0x00, 0x56, 0x18, 
0x00, 0x58, 0x18, 0x00, 0x5A, 0x18, 0x00, 0x5B, 0x18, 0x00, 0x61, 0x18, 0x00, 0x46, 0x18, 0x00, 
0x45, 0x18, 0x00, 0x43, 0x18, 0x00, 0x41, 0x18, 0x00, 0x3B, 0x18, 0x00, 0x3A, 0x18, 0x00, 0x38, 
0x18, 0x00, 0x36, 0x60, 0x0C, 0x8F, 0x00, 0x00, 0x60, 0x80, 0x00, 0x60, 0x10, 0x0C, 0x1F, 0x00, 
0x00, 0x9F, 0x00, 0x00, 0x60, 0x90, 0x00, 0x60, 0x20, 0x0C, 0x1F, 0x00, 0x00, 0xAF, 0x00, 0x00, 
0x60, 0xA0, 0x00, 0x60, 0x30, 0x0C, 0x1F, 0x00, 0x00, 0xBF, 0x00, 0x00, 0x60, 0xB0, 0x00, 0x60, 
0x40, 0x0C, 0x0F, 0x00, 0x00, 0x46, 0x18, 0x00, 0x48, 0x18, 0x00, 0x4A, 0x18, 0x00, 0x4B, 0x18, 
0x00, 0x51, 0x18, 0x00, 0x53, 0x18, 0x00, 0x55, 0x18, 0x00, 0x56, 0x18, 0x00, 0x58, 0x18, 0x00, 
0x5A, 0x18, 0x00, 0x5B, 0x18, 0x00, 0x61, 0x18, 0x00, 0x46, 0x18, 0x00, 0x45, 0x18, 0x00, 0x43, 
0x18, 0x00, 0x41, 0x18, 0x00, 0x3B, 0x18, 0x00, 0x3A, 0x18, 0x00, 0x38, 0x18, 0x00, 0x36, 0x60, 
0x0C, 0x8F, 0x00, 0x00, 0x60, 0xC0, 0x00, 0x60, 0x50, 0x0C, 0x1F, 0x00, 0x00, 0x9F, 0x00, 0x00, 
0x60, 0xD0, 0x00, 0x60, 0x60, 0x0C, 0x1F, 0x00, 0x00, 0xAF, 0x00, 0x00, 0x60, 0xE0, 0x00, 0x60, 
0x70, 0x0C, 0x1F, 0x00, 0x00, 0xBF, 0x00, 0x00, 0x60, 0xF0, 0x0C, 0xF0, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x10, 0x00, 0x90, 0x00, 0x80, 0x80, 0xC0, 0x2F, 0x10, 0x5D, 0x50, 0x80, 0xF1, 0x03, 0x20, 0x12, 
0xC0, 0xAF, 0xF0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x20, 0x0A, 0x00, 0x10, 0x00, 0x0D, 0x00, 0x02, 0x40, 0x0D, 0x00, 0xFF, 0x68, 0x0D, 0x00, 
0x10, 0x00, 0x00, 0x0E, 0x00, 0x00, 0x1E, 0x44, 0x00, 0x2E, 0x06, 0x00, 0x60, 0x00, 0x60, 0x0F, 
0x00, 0x00, 0x56, 0x18, 0x00, 0x57, 0x18, 0x00, 0x58, 0x18, 0x00, 0x59, 0x18, 0x00, 0x5A, 0x18, 
0x00, 0x5B, 0x18, 0x00, 0x5C, 0x18, 0x00, 0x61, 0x18, 0x00, 0x62, 0x18, 0x00, 0x63, 0x18, 0x00, 
0x64, 0x18, 0x00, 0x65, 0x18, 0x00, 0x66, 0x60, 0x00, 0x41, 0x18, 0x00, 0x3C, 0x18, 0x00, 0x3B, 
0x18, 0x00, 0x3A, 0x18, 0x00, 0x39, 0x18, 0x00, 0x38, 0x18, 0x00, 0x37, 0x18, 0x00, 0x36, 0x18, 
0x00, 0x35, 0x18, 0x00, 0x34, 0x18, 0x00, 0x33, 0x18, 0x00, 0x32, 0x18, 0x00, 0x31, 0x18, 0x00, 
0x2C, 0x18, 0x00, 0x2B, 0x18, 0x00, 0x2A, 0x18, 0x00, 0x29, 0x18, 0x00, 0x28, 0x18, 0x00, 0x27, 
0x18, 0x00, 0x26, 0x18, 0x00, 0x25, 0x18, 0x00, 0x24, 0x18, 0x00, 0x23, 0x18, 0x00, 0x22, 0x18, 
0x00, 0x21, 0x48, 0x0C, 0x8F, 0x00, 0x00, 0x60, 0x80, 0x00, 0x60, 0x10, 0x0C, 0x1F, 0x00, 0x00, 
0x9F, 0x00, 0x00, 0x60, 0x90, 0x00, 0x60, 0x20, 0x0C, 0x1F, 0x00, 0x00, 0xAF, 0x00, 0x00, 0x60, 
0xA0, 0x00, 0x60, 0x30, 0x0C, 0x1F, 0x00, 0x00, 0xBF, 0x00, 0x00, 0x60, 0xB0, 0x00, 0x60, 0x40, 
0x0C, 0x0F, 0x00, 0x00, 0x56, 0x18, 0x00, 0x57, 0x18, 0x00, 0x58, 0x18, 0x00, 0x59, 0x18, 0x00, 
0x5A, 0x18, 0x00, 0x5B, 0x18, 0x00, 0x5C, 0x18, 0x00, 0x61, 0x18, 0x00, 0x62, 0x18, 0x00, 0x63, 
0x18, 0x00, 0x64, 0x18, 0x00, 0x65, 0x18, 0x00, 0x66, 0x60, 0x00, 0x41, 0x18, 0x00, 0x3C, 0x18, 
0x00, 0x3B, 0x18, 0x00, 0x3A, 0x18, 0x00, 0x39, 0x18, 0x00, 0x38, 0x18, 0x00, 0x37, 0x18, 0x00, 
0x36, 0x18, 0x00, 0x35, 0x18, 0x00, 0x34, 0x18, 0x00, 0x33, 0x18, 0x00, 0x32, 0x18, 0x00, 0x31, 
0x18, 0x00, 0x2C, 0x18, 0x00, 0x2B, 0x18, 0x00, 0x2A, 0x18, 0x00, 0x29, 0x18, 0x00, 0x28, 0x18, 
0x00, 0x27, 0x18, 0x00, 0x26, 0x18, 0x00, 0x25, 0x18, 0x00, 0x24, 0x18, 0x00, 0x23, 0x18, 0x00, 
0x22, 0x18, 0x00, 0x21, 0x48, 0x0C, 0x8F, 0x00, 0x00, 0x60, 0xC0, 0x00, 0x60, 0x50, 0x0C, 0x1F, 
0x00, 0x00, 0x9F, 0x00, 0x00, 0x60, 0xD0, 0x00, 0x60, 0x60, 0x0C, 0x1F, 0x00, 0x00, 0xAF, 0x00, 
0x00, 0x60, 0xE0, 0x00, 0x60, 0x70, 0x0C, 0x1F, 0x00, 0x00, 0xBF, 0x00, 0x00, 0x60, 0xF0, 0x0C, 
0xF0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x10, 0x00, 0x00, 0x60, 0x00, 0x00, 0x60, 0x80, 0x60, 0x10, 0x00, 0x21, 0xF0, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x10, 0x00, 0x90, 0x00, 0x80, 0x80, 0xC0, 0x2F, 0x10, 0x5D, 0x50, 0x80, 0xF1, 0x03, 0x20, 0x21, 
0xC0, 0xAF, 0xF0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x20, 0x0E, 0x00, 0x10, 0x60, 0x0E, 0x00, 0x02, 0x20, 0x11, 0x00, 0xFF, 0x48, 0x11, 0x00, 
0x10, 0x00, 0x00, 0x0E, 0x00, 0x00, 0x1E, 0x44, 0x00, 0x2E, 0x06, 0x00, 0x60, 0x00, 0x00, 0x60, 
0x80, 0x60, 0x10, 0x00, 0x21, 0xF0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x10, 0x00, 0x00, 0x60, 0x00, 0x60, 0x0F, 0x00, 0x00, 0x56, 0x18, 0x00, 0x57, 0x18, 0x00, 0x58, 
0x18, 0x00, 0x59, 0x18, 0x00, 0x5A, 0x18, 0x00, 0x5B, 0x18, 0x00, 0x5C, 0x18, 0x00, 0x61, 0x18, 
0x00, 0x62, 0x18, 0x00, 0x63, 0x18, 0x00, 0x64, 0x18, 0x00, 0x65, 0x18, 0x00, 0x66, 0x60, 0x00, 
0x41, 0x18, 0x00, 0x3C, 0x18, 0x00, 0x3B, 0x18, 0x00, 0x3A, 0x18, 0x00, 0x39, 0x18, 0x00, 0x38, 
0x18, 0x00, 0x37, 0x18, 0x00, 0x36, 0x18, 0x00, 0x35, 0x18, 0x00, 0x34, 0x18, 0x00, 0x33, 0x18, 
0x00, 0x32, 0x18, 0x00, 0x31, 0x18, 0x00, 0x2C, 0x18, 0x00, 0x2B, 0x18, 0x00, 0x2A, 0x18, 0x00, 
0x29, 0x18, 0x00, 0x28, 0x18, 0x00, 0x27, 0x18, 0x00, 0x26, 0x18, 0x00, 0x25, 0x18, 0x00, 0x24, 
0x18, 0x00, 0x23, 0x18, 0x00, 0x22, 0x18, 0x00, 0x21, 0x48, 0x0C, 0x8F, 0x00, 0x00, 0x60, 0x80, 
0x00, 0x60, 0x10, 0x0C, 0x1F, 0x00, 0x00, 0x9F, 0x00, 0x00, 0x60, 0x90, 0x00, 0x60, 0x20, 0x0C, 
0x1F, 0x00, 0x00, 0xAF, 0x00, 0x00, 0x60, 0xA0, 0x00, 0x60, 0x30, 0x0C, 0x1F, 0x00, 0x00, 0xBF, 
0x00, 0x00, 0x60, 0xB0, 0x00, 0x60, 0x40, 0x0C, 0x0F, 0x00, 0x00, 0x56, 0x18, 0x00, 0x57, 0x18, 
0x00, 0x58, 0x18, 0x00, 0x59, 0x18, 0x00, 0x5A, 0x18, 0x00, 0x5B, 0x18, 0x00, 0x5C, 0x18, 0x00, 
0x61, 0x18, 0x00, 0x62, 0x18, 0x00, 0x63, 0x18, 0x00, 0x64, 0x18, 0x00, 0x65, 0x18, 0x00, 0x66, 
0x60, 0x00, 0x41, 0x18, 0x00, 0x3C, 0x18, 0x00, 0x3B, 0x18, 0x00, 0x3A, 0x18, 0x00, 0x39, 0x18, 
0x00, 0x38, 0x18, 0x00, 0x37, 0x18, 0x00, 0x36, 0x18, 0x00, 0x35, 0x18, 0x00, 0x34, 0x18, 0x00, 
0x33, 0x18, 0x00, 0x32, 0x18, 0x00, 0x31, 0x18, 0x00, 0x2C, 0x18, 0x00, 0x2B, 0x18, 0x00, 0x2A, 
0x18, 0x00, 0x29, 0x18, 0x00, 0x28, 0x18, 0x00, 0x27, 0x18, 0x00, 0x26, 0x18, 0x00, 0x25, 0x18, 
0x00, 0x24, 0x18, 0x00, 0x23, 0x18, 0x00, 0x22, 0x18, 0x00, 0x21, 0x48, 0x0C, 0x8F, 0x00, 0x00, 
0x60, 0xC0, 0x00, 0x60, 0x50, 0x0C, 0x1F, 0x00, 0x00, 0x9F, 0x00, 0x00, 0x60, 0xD0, 0x00, 0x60, 
0x60, 0x0C, 0x1F, 0x00, 0x00, 0xAF, 0x00, 0x00, 0x60, 0xE0, 0x00, 0x60, 0x70, 0x0C, 0x1F, 0x00, 
0x00, 0xBF, 0x00, 0x00, 0x60, 0xF0, 0x0C, 0xF0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x10, 0x00, 0x90, 0x00, 0x80, 0x80, 0xC0, 0x2F, 0x10, 0x5D, 0x50, 0x80, 0xF1, 0x03, 0x20, 0x21, 
0xC0, 0xAF, 0xF0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x20, 0x12, 0x00, 0x10, 0x60, 0x12, 0x00, 0x02, 0xA0, 0x12, 0x00, 0xFF, 0xE0, 0x16, 0x00, 
0x10, 0x00, 0x00, 0x0E, 0x00, 0x00, 0x1E, 0x44, 0x00, 0x2E, 0x01, 0x00, 0x60, 0x00, 0x00, 0x60, 
0x80, 0x60, 0x10, 0x80, 0x2E, 0xF0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x10, 0x00, 0x00, 0x60, 0x00, 0x00, 0x60, 0x80, 0x60, 0x10, 0x80, 0x2E, 0xF0, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x10, 0x00, 0x90, 0x00, 0x80, 0x80, 0xC0, 0x26, 0x10, 0x5D, 0x50, 0x00, 0xF1, 0x20, 0x20, 0x01, 
0x60, 0x00, 0x10, 0x00, 0x10, 0x03, 0x60, 0x00, 0x10, 0x00, 0x10, 0x5D, 0x60, 0x80, 0x10, 0x20, 
0x20, 0x01, 0x50, 0x80, 0x50, 0x50, 0x10, 0x18, 0x50, 0xD0, 0xF1, 0x48, 0x50, 0x01, 0xF1, 0x20, 
0x20, 0x01, 0x60, 0x01, 0x10, 0x00, 0x10, 0x03, 0x60, 0x01, 0x10, 0x00, 0x10, 0x5D, 0x60, 0x81, 
0x10, 0x20, 0x20, 0x01, 0x50, 0x81, 0x50, 0x50, 0x10, 0x18, 0x50, 0xD0, 0xF1, 0x48, 0x50, 0x10, 
0xF1, 0x20, 0x20, 0x01, 0x60, 0x10, 0x10, 0x00, 0x10, 0x03, 0x60, 0x10, 0x10, 0x00, 0x10, 0x5D, 
0x60, 0x90, 0x10, 0x20, 0x20, 0x01, 0x50, 0x90, 0x50, 0x50, 0x10, 0x18, 0x50, 0xD0, 0xF1, 0x48, 
0x50, 0x11, 0xF1, 0x20, 0x20, 0x01, 0x60, 0x11, 0x10, 0x00, 0x10, 0x03, 0x60, 0x11, 0x10, 0x00, 
0x10, 0x5D, 0x60, 0x91, 0x10, 0x20, 0x20, 0x01, 0x50, 0x91, 0x50, 0x50, 0x10, 0x18, 0x50, 0xD0, 
0xF1, 0x48, 0x50, 0x20, 0xF1, 0x20, 0x20, 0x01, 0x60, 0x20, 0x10, 0x00, 0x10, 0x03, 0x60, 0x20, 
0x10, 0x00, 0x10, 0x5D, 0x60, 0xA0, 0x10, 0x20, 0x20, 0x01, 0x50, 0xA0, 0x50, 0x50, 0x10, 0x18, 
0x50, 0xD0, 0xF1, 0x48, 0x50, 0x21, 0xF1, 0x20, 0x20, 0x01, 0x60, 0x21, 0x10, 0x00, 0x10, 0x03, 
0x60, 0x21, 0x10, 0x00, 0x10, 0x5D, 0x60, 0xA1, 0x10, 0x20, 0x20, 0x01, 0x50, 0xA1, 0x50, 0x50, 
0x10, 0x18, 0x50, 0xD0, 0xF1, 0x48, 0x50, 0x30, 0xF1, 0x20, 0x20, 0x01, 0x60, 0x30, 0x10, 0x00, 
0x10, 0x03, 0x60, 0x30, 0x10, 0x00, 0x10, 0x5D, 0x60, 0xB0, 0x10, 0x20, 0x20, 0x01, 0x50, 0xB0, 
0x50, 0x50, 0x10, 0x18, 0x50, 0xD0, 0xF1, 0x48, 0x50, 0x31, 0xF1, 0x20, 0x20, 0x01, 0x60, 0x31, 
0x10, 0x00, 0x10, 0x03, 0x60, 0x31, 0x10, 0x00, 0x10, 0x5D, 0x60, 0xB1, 0x10, 0x20, 0x20, 0x01, 
0x50, 0xB1, 0x50, 0x50, 0x10, 0x18, 0x50, 0xD0, 0xF1, 0x48, 0x50, 0x40, 0xF1, 0x20, 0x20, 0x01, 
0x60, 0x40, 0x10, 0x00, 0x10, 0x03, 0x60, 0x40, 0x10, 0x00, 0x10, 0x5D, 0x60, 0xC0, 0x10, 0x20, 
0x20, 0x01, 0x50, 0xC0, 0x50, 0x50, 0x10, 0x18, 0x50, 0xD0, 0xF1, 0x48, 0x50, 0x41, 0xF1, 0x20, 
0x20, 0x01, 0x60, 0x41, 0x10, 0x00, 0x10, 0x03, 0x60, 0x41, 0x10, 0x00, 0x10, 0x5D, 0x60, 0xC1, 
0x10, 0x20, 0x20, 0x01, 0x50, 0xC1, 0x50, 0x50, 0x10, 0x18, 0x50, 0xD0, 0xF1, 0x48, 0x50, 0x50, 
0xF1, 0x20, 0x20, 0x01, 0x60, 0x50, 0x10, 0x00, 0x10, 0x03, 0x60, 0x50, 0x10, 0x00, 0x10, 0x5D, 
0x60, 0xD0, 0x10, 0x20, 0x20, 0x01, 0x50, 0xD0, 0x50, 0x50, 0x10, 0x18, 0x50, 0xD0, 0xF1, 0x48, 
0x50, 0x51, 0xF1, 0x20, 0x20, 0x01, 0x60, 0x51, 0x10, 0x00, 0x10, 0x03, 0x60, 0x51, 0x10, 0x00, 
0x10, 0x5D, 0x60, 0xD1, 0x10, 0x20, 0x20, 0x01, 0x50, 0xD1, 0x50, 0x50, 0x10, 0x18, 0x50, 0xD0, 
0xF1, 0x48, 0x50, 0x60, 0xF1, 0x20, 0x20, 0x01, 0x60, 0x60, 0x10, 0x00, 0x10, 0x03, 0x60, 0x60, 
0x10, 0x00, 0x10, 0x5D, 0x60, 0xE0, 0x10, 0x20, 0x20, 0x01, 0x50, 0xE0, 0x50, 0x50, 0x10, 0x18, 
0x50, 0xD0, 0xF1, 0x48, 0x50, 0x61, 0xF1, 0x20, 0x20, 0x01, 0x60, 0x61, 0x10, 0x00, 0x10, 0x03, 
0x60, 0x61, 0x10, 0x00, 0x10, 0x5D, 0x60, 0xE1, 0x10, 0x20, 0x20, 0x01, 0x50, 0xE1, 0x50, 0x50, 
0x10, 0x18, 0x50, 0xD0, 0xF1, 0x48, 0x50, 0x70, 0xF1, 0xD8, 0x60, 0x70, 0x10, 0x00, 0x10, 0x03, 
0x60, 0x70, 0x10, 0x00, 0x10, 0x45, 0x60, 0xF0, 0x10, 0xF0, 0x50, 0xF0, 0xF1, 0x30, 0x50, 0x71, 
0xF1, 0xD8, 0x60, 0x71, 0x10, 0x00, 0x10, 0x03, 0x60, 0x71, 0x10, 0x00, 0x10, 0x45, 0x60, 0xF1, 
0x10, 0xD8, 0x50, 0xF1, 0x50, 0x70, 0x10, 0x18, 0x50, 0xF0, 0xF1, 0x33, 0xC0, 0xA6, 0xF0, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x20, 0x18, 0x00, 0x10, 0x60, 0x18, 0x00, 0x02, 0xA0, 0x18, 0x00, 0xFF, 0xC8, 0x1B, 0x00, 
0x10, 0x00, 0x00, 0x0E, 0x00, 0x00, 0x1E, 0x44, 0x00, 0x2E, 0x01, 0x00, 0x60, 0x00, 0x00, 0x60, 
0x80, 0x60, 0x10, 0x00, 0x48, 0xF0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x10, 0x00, 0x00, 0x60, 0x00, 0x00, 0x60, 0x80, 0x60, 0x10, 0x00, 0x48, 0xF0, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x10, 0x00, 0x90, 0x00, 0xC0, 0x26, 0x50, 0x80, 0x10, 0x60, 0x90, 0x00, 0x01, 0x60, 0x11, 0x60, 
0x21, 0x60, 0x31, 0x60, 0x41, 0x60, 0x51, 0x60, 0x61, 0x60, 0x71, 0x60, 0x81, 0x60, 0x91, 0x60, 
0xA1, 0x60, 0xB1, 0x60, 0xC1, 0x60, 0xD1, 0x60, 0xE1, 0x60, 0xF1, 0x60, 0x06, 0x60, 0x16, 0x60, 
0x26, 0x60, 0x36, 0x60, 0x46, 0x60, 0x56, 0x60, 0x66, 0x60, 0x76, 0x60, 0x86, 0x60, 0x96, 0x60, 
0xA6, 0x60, 0xB6, 0x60, 0xC6, 0x60, 0xD6, 0x60, 0xE6, 0x60, 0xF6, 0x60, 0x0B, 0x60, 0x1B, 0x60, 
0x2B, 0x60, 0x3B, 0x60, 0x4B, 0x60, 0x5B, 0x60, 0x6B, 0x60, 0x7B, 0x60, 0x8B, 0x60, 0x9B, 0x60, 
0xAB, 0x60, 0xBB, 0x60, 0xCB, 0x60, 0xDB, 0x60, 0xEB, 0x60, 0xFB, 0x60, 0x04, 0x60, 0x14, 0x60, 
0x24, 0x60, 0x34, 0x60, 0x44, 0x60, 0x54, 0x60, 0x64, 0x60, 0x74, 0x60, 0x84, 0x60, 0x94, 0x60, 
0xA4, 0x60, 0xB4, 0x60, 0xC4, 0x60, 0xD4, 0x60, 0xE4, 0x60, 0xF4, 0x60, 0x09, 0x60, 0x19, 0x60, 
0x29, 0x60, 0x39, 0x60, 0x49, 0x60, 0x59, 0x60, 0x69, 0x60, 0x79, 0x60, 0x89, 0x60, 0x99, 0x60, 
0xA9, 0x60, 0xB9, 0x60, 0xC9, 0x60, 0xD9, 0x60, 0xE9, 0x60, 0xF9, 0x60, 0x02, 0x60, 0x12, 0x60, 
0x22, 0x60, 0x32, 0x60, 0x42, 0x60, 0x52, 0x60, 0x62, 0x60, 0x72, 0x60, 0x82, 0x60, 0x92, 0x60, 
0xA2, 0x60, 0xB2, 0x60, 0xC2, 0x60, 0xD2, 0x60, 0xE2, 0x60, 0xF2, 0x60, 0x07, 0x60, 0x17, 0x60, 
0x27, 0x60, 0x37, 0x60, 0x47, 0x60, 0x57, 0x60, 0x67, 0x60, 0x77, 0x60, 0x87, 0x60, 0x97, 0x60, 
0xA7, 0x60, 0xB7, 0x60, 0xC7, 0x60, 0xD7, 0x60, 0xE7, 0x60, 0xF7, 0x60, 0x0C, 0x60, 0x1C, 0x60, 
0x2C, 0x60, 0x3C, 0x60, 0x4C, 0x60, 0x5C, 0x60, 0x6C, 0x60, 0x7C, 0x60, 0x8C, 0x60, 0x9C, 0x60, 
0xAC, 0x60, 0xBC, 0x60, 0xCC, 0x60, 0xDC, 0x60, 0xEC, 0x60, 0xFC, 0x60, 0x05, 0x60, 0x15, 0x60, 
0x25, 0x60, 0x35, 0x60, 0x45, 0x60, 0x55, 0x60, 0x65, 0x60, 0x75, 0x60, 0x85, 0x60, 0x95, 0x60, 
0xA5, 0x60, 0xB5, 0x60, 0xC5, 0x60, 0xD5, 0x60, 0xE5, 0x60, 0xF5, 0x60, 0x0A, 0x60, 0x1A, 0x60, 
0x2A, 0x60, 0x3A, 0x60, 0x4A, 0x60, 0x5A, 0x60, 0x6A, 0x60, 0x7A, 0x60, 0x8A, 0x60, 0x9A, 0x60, 
0xAA, 0x60, 0xBA, 0x60, 0xCA, 0x60, 0xDA, 0x60, 0xEA, 0x60, 0xFA, 0x60, 0x03, 0x60, 0x13, 0x60, 
0x23, 0x60, 0x33, 0x60, 0x43, 0x60, 0x53, 0x60, 0x63, 0x60, 0x73, 0x60, 0x83, 0x60, 0x93, 0x60, 
0xA3, 0x60, 0xB3, 0x60, 0xC3, 0x60, 0xD3, 0x60, 0xE3, 0x60, 0xF3, 0x60, 0x08, 0x60, 0x18, 0x60, 
0x28, 0x60, 0x38, 0x60, 0x48, 0x60, 0x58, 0x60, 0x68, 0x60, 0x78, 0x60, 0x88, 0x60, 0x98, 0x60, 
0xA8, 0x60, 0xB8, 0x60, 0xC8, 0x60, 0xD8, 0x60, 0xE8, 0x60, 0xF8, 0x60, 0x50, 0x80, 0xF1, 0x03, 
0xC0, 0xA6, 0xF0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x20, 0x1E, 0x00, 0x10, 0xE0, 0x1E, 0x00, 0x02, 0x80, 0x1F, 0x00, 0xFF, 0x9C, 0x1F, 0x00, 
0x10, 0x00, 0x00, 0x0E, 0x00, 0x00, 0x1E, 0x44, 0x00, 0x2E, 0x01, 0x00, 0x60, 0x00, 0x0C, 0x51, 
0x06, 0x00, 0x53, 0x06, 0x00, 0x55, 0x06, 0x00, 0x56, 0x06, 0x00, 0x58, 0x0C, 0x00, 0x61, 0x0C, 
0x00, 0x5A, 0x18, 0x00, 0x56, 0x18, 0x00, 0x58, 0x06, 0x00, 0x51, 0x06, 0x00, 0x53, 0x06, 0x00, 
0x55, 0x06, 0x00, 0x56, 0x06, 0x00, 0x58, 0x06, 0x00, 0x5A, 0x06, 0x00, 0x5C, 0x06, 0x00, 0x61, 
0x0C, 0x24, 0x60, 0x80, 0x00, 0xF0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x10, 0x00, 0x00, 0x60, 0x00, 0x3C, 0x4A, 0x06, 0x00, 0x48, 0x06, 0x00, 0x46, 0x06, 0x00, 0x45, 
0x06, 0x00, 0x43, 0x0C, 0x00, 0x48, 0x0C, 0x00, 0x41, 0x06, 0x00, 0x45, 0x06, 0x00, 0x46, 0x06, 
0x00, 0x48, 0x06, 0x00, 0x4A, 0x06, 0x00, 0x4C, 0x06, 0x00, 0x51, 0x06, 0x00, 0x53, 0x06, 0x00, 
0x55, 0x0C, 0x24, 0x60, 0x80, 0x00, 0xF0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x10, 0x00, 0x90, 0x00, 0xC0, 0x2A, 0x50, 0x80, 0xF1, 0xCC, 0xC0, 0xAA, 0xF0, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x23, 0x83, 0x93, 0x06, 0x47, 0x83, 0xAB, 0x02, 
0x63, 0x27, 0x4B, 0x27, 0x47, 0x93, 0x2B, 0x83, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00

// exactly 4 kB
};

unsigned int addressPointer = 0;                              //  Data Address
unsigned int addressPointerxxxx;                              //    Bits 4-7 of Data Address
unsigned int addressPointeryyyy;                              //    Bits 8-11 of Data Address
unsigned int addressPointerzzzz;                              //    Bits 12-15 of Data Address

bool deviceCode = false;                                      //  True if proper device code received.  Otherwise false
bool midByte = false;                                         //  True if only first half of byte read (PT-50 does this)

byte dataByte;                                                //  Data to be sent
byte dataByteLow;                                             //    Bits 0-3 of Data
byte dataByteHigh;                                            //    Bits 4-7 of Data

void setup() {
  DDRB = B00000000;                                           //  OP, CE (pins 10, 11) set as INPUTs (along with the rest of the port)
  PORTB = B11110011;                                          //  Disable PULL-UPs for OP, CE
  
  DDRD = B00000000;                                           //  CK1, CK2, DATA1-DATA4 (pins 2, 3, 4-7) set as INPUTs
  PORTD = B00000011;                                          //  Disable PULL-UPs for CK1, CK2, DATA lines
  
  noInterrupts();                                             //  Disable Interrupts--Arduino's background interrupts impact the timing
}

void loop() {                                                 //  The main loop is simply the emulator acting on changes in CE & OP
  while ((PIND & B00001100) != B00001000) { }                 //  Wait for CK1
  while ((PIND & B00001100) != B00000100) { }                 //  Wait for CK2
  while ((PIND & B00001100) != B00001100) { }                 //  Wait for after CK2 (CE & OP should arrive at CK2, but timing isn't always perfect)
  switch (PINB & B00001100) {                                 //  Test if CE or OP are LOW

    // ================ ADDRESS OPERATION ================
    case B00000000:                                           //  OP and CE LOW

      while ((PIND & B00001100) != B00001000) { }             //  Wait for CK1  Wait to make sure data from keyboard is on pins
      while ((PIND & B00001100) != B00000100) { }             //  Wait for CK2
      while ((PIND & B00001100) != B00001000) { }             //  Wait for CK1  READ code is on pins, but ignore--it's always 0000

      while ((PIND & B00001100) != B00000100) { }             //  Wait for CK2  Advance to xxxx
      while ((PIND & B00001100) != B00001000) { }             //  Wait for CK1
      while ((PIND & B00001100) != B00000100) { }             //  Wait for CK2
      while ((PIND & B00001100) != B00001000) { }             //  Wait for CK1
      addressPointerxxxx = PIND;                              //  Capture xxxx

      while ((PIND & B00001100) != B00000100) { }             //  Wait for CK2  Advance to yyyy
      while ((PIND & B00001100) != B00001000) { }             //  Wait for CK1
      while ((PIND & B00001100) != B00000100) { }             //  Wait for CK2
      while ((PIND & B00001100) != B00001000) { }             //  Wait for CK1
      addressPointeryyyy = PIND;                              //  Capture yyyy

      while ((PIND & B00001100) != B00000100) { }             //  Wait for CK2  Advance to zzzz
      while ((PIND & B00001100) != B00001000) { }             //  Wait for CK1
      while ((PIND & B00001100) != B00000100) { }             //  Wait for CK2
      while ((PIND & B00001100) != B00001000) { }             //  Wait for CK1
      addressPointerzzzz = PIND;                              //  Capture zzzz

      while ((PIND & B00001100) != B00000100) { }             //  Wait for CK2  Advance to Device code
      while ((PIND & B00001100) != B00001000) { }             //  Wait for CK1
      while ((PIND & B00001100) != B00000100) { }             //  Wait for CK2
      while ((PIND & B00001100) != B00001000) { }             //  Wait for CK1
      if ((~PIND & B11110000) == 0xC0) {                      //  Device Code.  Set deviceCode to true if correct code received
        deviceCode = true;
      }
      else {
        deviceCode = false;                                   //  Set deviceCode to false if invalid code received (this will skip future DATA OUT ops)
      }
      midByte = false;                                        //  New addressing op cancels any mid-byte read (should not be necessary)
      
      while ((PINB & B00001100) != B00001100) { }             //  Wait for CE and OP to go HIGH
      
      if (deviceCode == true) {                               //  Only set Address Pointer if proper Device Code was received
        addressPointer = 0;                                   //  Reset Address Pointer to $0000
        addressPointer += addressPointerxxxx | 0x0F;          //   0000  0000 ~xxxx ~0000 (everything is ~ because of negative logic)
        addressPointer += (addressPointeryyyy & 0xF0) << 4;   //   0000 ~yyyy ~xxxx ~0000
        addressPointer += (addressPointerzzzz & 0xF0) << 8;   //  ~zzzz ~yyyy ~xxxx ~0000
        addressPointer = ~addressPointer >> 1;                //  Invert and shift right (divide by 2) to convert Word address to Byte address
        dataByte = pgm_read_byte(&dataROM[addressPointer]);   //  Pre-load the data byte (bbbbaaaa) from address
        dataByteLow = ~dataByte << 4;                         //    Invert for negative logic (should now be ----aaaa), shift to high word (aaaa0000)
        dataByteHigh = ~dataByte & B11110000;                 //    Invert for negative logic (should now be bbbb----), zero out bottom (bbbb0000)
      }
    break;

    // ================ DATA OUT OPERATION ================
    case B00000100:                                               //  Only CE LOW
      if (deviceCode == true) {                                   //  Read data out unless proper Device Code was not received
        do {                                                      //  Keep reading out more data so long as CE remains low
          if (midByte == false) {                                 //  skip low half of byte if mid-byte set
            while ((PIND & B00001100) != B00001000) { }           //  Wait for CK1
            PORTD = dataByteLow;                                  //  Write lower portion of dataByte to pins
            DDRD = B11110000;                                     //  Set DATA lines to OUTPUTs
            while ((PIND & B00001100) != B00000100) { }           //  Wait for CK2
            while ((PIND & B00001100) != B00001000) { }           //  Wait for CK1
            while ((PIND & B00001100) != B00000100) { }           //  Wait for CK2
            while ((PIND & B00001100) != B00001100) { }           //  Wait for after CK2 (If CE and CK2 timing isn't perfect)
          }
          if ((PINB & B00001000) == B00000000) {                  //  If CE still LOW, continue with upper portion of dataByte and Address Increment
            while ((PIND & B00001100) != B00001000) { }           //  Wait for CK1
            PORTD = dataByteHigh;                                 //  Write upper portion of dataByte to pins
            DDRD = B11110000;                                     //  Set DATA lines to OUTPUTs
            while ((PIND & B00001100) != B00000100) { }           //  Wait for CK2
            while ((PIND & B00001100) != B00001000) { }           //  Wait for CK1
            while ((PIND & B00001100) != B00000100) { }           //  Wait for CK2
            while ((PIND & B00001100) != B00001100) { }           //  Wait for after CK2 (If CE and CK2 timing isn't perfect)

            midByte = false;                                      //  Clear midByte flag

            addressPointer += 1;                                  //  Increment addressPointer
            dataByte = pgm_read_byte(&dataROM[addressPointer]);   //  Pre-load the data byte (bbbbaaaa) from address
            dataByteLow = ~dataByte << 4;                         //    Invert for negative logic (should now be ----aaaa), shift to high word (aaaa0000)
            dataByteHigh = ~dataByte & B11110000;                 //    Invert for negative logic (should now be bbbb----), zero out bottom (bbbb0000)
          }
          else {
            midByte = true;                                       //  If keyboard ceases DATA OUT before second half of byte (PT-50 does this)
          }
          
        } while ((PINB & B00001100) != B00001100);                //  Continue sequential Data Read if CE remains LOW.

        DDRD = B00000000;                                         //  Return DATA lines to INPUTs
        PORTD = B00000011;                                        //  Disable PULL-UPs
      }
    break;
  }
}


/*  ======/ Special Thanks to /====================================================================

    http://www.crumblenet.co.uk/keyb/rpacks.html
      Tim has studied the many ROM packs of several different Casio and other keyboards.

    http://www.pisi.com.pl/piotr433/hardware.htm
      Piotr has analyzed the SRAM of the Casio FX-700P a chip similar to Casio ROM

    https://seanriddle.com/m5268.html
      Sean dumped the contents of RO-251, starting me on this adventure.
  
    ===============================================================================================  */
